﻿using System;
using System.IO;
using System.Linq;
using System.Text;
using System.Xml.Linq;
using System.Xml.XPath;
using System.Xml;
using System.Xml.Schema;
using OpenSSL.Core;
using OpenSSL.Crypto;
using OpenSSL.X509;
using Org.BouncyCastle.Crypto.Digests;
using Org.BouncyCastle.Crypto.Engines;
using Org.BouncyCastle.Crypto.Parameters;
using Org.BouncyCastle.Crypto.Signers;
using Org.BouncyCastle.Math;
using Org.BouncyCastle.Security;

//TODO: verifyTimestamp function, rozsirit verifyCertificate o platnost, CRL

namespace DocasneUloziskoHlasov
{
    class Program
    {
        static string configFile = "duhConfig.xml";

        static string voteInputSchemaFile;
        static string voteDecryptedSchemaFile;
        static string voteDecryptedWithTimeStampSchemaFile;
        static int totalSisCertificateNumber;
        static string caRootCertPublicKeyFile;
        static string activeCertPass;
        static string temporaryCertificateFile;
        static int[] activeCertificatesSN;

        static CryptoKey temporaryCertificate;
        static CryptoKey caRootCertPublicKey;
        static XDocument voteInputSchema;
        static XDocument voteDecryptedSchema;
        static XDocument voteDecryptedWithTimeStampSchema;

        private static string getRandomFile(string path, string extension)
        {
            string file = null;
            if (!string.IsNullOrEmpty(path))
            {
                var extensions = new string[] { extension };
                try
                {
                    var di = new DirectoryInfo(path);
                    var rgFiles = di.GetFiles("*.*").Where(f => extensions.Contains(f.Extension.ToLower()));
                    SecureRandom R = new SecureRandom();
                    file = rgFiles.ElementAt(R.Next(0, rgFiles.Count())).FullName;
                }
                catch { }
            }
            return file;
        }

        public static BigInteger opensslBigToBigInteger(String opensslInt)
        {
            try
            {
                String hexInt = opensslInt.Replace(":", "");
                hexInt = hexInt.Replace("\\s+", "");
                return new BigInteger(hexInt, 16);
            }
            catch
            {
                return null;
            }
        }   

        static XDocument decipherVote(XDocument vote, CryptoKey cert)
        {
            //predpoklada na vstupe validny hlas
            byte[] secondEnvelopeEnc = Convert.FromBase64String(vote.XPathSelectElement("/voteBallotThirdEnvelope/voteBallotEncrypted").Value);
            byte[] secondEnvelopeEncKey = Convert.FromBase64String(vote.XPathSelectElement("/voteBallotThirdEnvelope/voteBallotEncryptionKey").Value);

            try
            {
                RSA thirdEnvelopeDecryption = cert.GetRSA();
                byte[] decData = thirdEnvelopeDecryption.PrivateDecrypt(secondEnvelopeEncKey, RSA.Padding.OAEP);

                CipherContext symetricCipher = new CipherContext(Cipher.AES_256_CBC);
                byte[] iv = new byte[8];
                Buffer.BlockCopy(decData, 0, iv, 0, 8);
                byte[] symKey = new byte[256];
                Buffer.BlockCopy(decData, 8, symKey, 0, 256);
                byte[] secondEnvelope = symetricCipher.Decrypt(secondEnvelopeEnc, symKey, iv);
                XDocument result = XDocument.Parse(Encoding.ASCII.GetString(secondEnvelope));
                if (!Validate_XML(result, voteDecryptedSchema))
                    return null;
                result.Root.LastNode.AddAfterSelf(vote.XPathSelectElement("/voteBallotThirdEnvelope/timestamp"));
                return result;
            }
            catch
            {
                return null;
            }
        }

        static bool verifyCertificate(X509Certificate inputCert)
        {
            //neuplne - overuje iba hash a podpis hashu
            //neoveruje cas a platnost voci CRL

            return inputCert.Verify(caRootCertPublicKey);
        }

        static bool verifyTimestamp(XDocument vote)
        {
            //TODO

            return true;
        }

        static bool verifyVote(XDocument vote)
        {
            // predpoklada na vstupe validny hlas
            X509Certificate voterSignCert;
            BIO voterCertBio = BIO.MemoryBuffer();
            voterCertBio.Write(Convert.FromBase64String(vote.XPathSelectElement("/voteBallotSecondEnvelope/voterSignCertificate").Value));
            try
            {
                voterSignCert = X509Certificate.FromDER(voterCertBio);
            }
            catch
            {
                return false;
            }

            // verify voter's certificate
            if (!verifyCertificate(voterSignCert))
                return false;

            // verify timestamp
            if (!verifyTimestamp(vote))
                return false;
            
            // verify signature value
            try
            {
                RSA rsa = voterSignCert.PublicKey.GetRSA();
                byte[] signatureValue = Convert.FromBase64String(vote.XPathSelectElement("/voteBallotSecondEnvelope/voteBallotSignature").Value);
                byte[] data = Convert.FromBase64String(vote.XPathSelectElement("/voteBallotSecondEnvelope/voteBallot").Value);

                RsaKeyParameters signKey = new RsaKeyParameters(false, opensslBigToBigInteger(rsa.PublicModulus.ToString()), opensslBigToBigInteger(rsa.PublicExponent.ToString()));
                PssSigner voterSignVerifier = new PssSigner(new RsaEngine(), new Sha512Digest());
                voterSignVerifier.Init(false, signKey);
                voterSignVerifier.BlockUpdate(data, 0, data.Length);
                if (!voterSignVerifier.VerifySignature(signatureValue))
                    return false;
            }
            catch
            {
                return false;
            }

            return true;
        }

        public static bool Validate_XML(XDocument validated_xml, XDocument xsd_schema)
        {
            bool no_errors = true;
            try
            {
                XmlSchemaSet schemas = new XmlSchemaSet();
                schemas.Add(null, XmlReader.Create(new StringReader(xsd_schema.ToString())));
                validated_xml.Validate(schemas, null, true);
            }
            catch
            {
                no_errors = false;
            }
            return no_errors;
        }

        static void Main(string[] args)
        {
            Console.WriteLine("Docasne Ulozisko Hlasov (DUH)");
            try
            {
                XDocument config = XDocument.Load(configFile);
                voteDecryptedSchemaFile = config.XPathSelectElement("/duhConfig/voteDecryptedSchemaFile").Value.Trim();
                voteDecryptedWithTimeStampSchemaFile = config.XPathSelectElement("/duhConfig/voteDecryptedWithTimeStampSchemaFile").Value.Trim();
                voteInputSchemaFile = config.XPathSelectElement("/duhConfig/voteInputSchemaFile").Value.Trim();
                totalSisCertificateNumber = int.Parse(config.XPathSelectElement("/duhConfig/totalSisCertificateNumber").Value.Trim());
                caRootCertPublicKeyFile = config.XPathSelectElement("/duhConfig/caRootCertPublicKeyFile").Value.Trim();

                temporaryCertificateFile = "Certs/" + totalSisCertificateNumber + ".pem";
            }
            catch
            {
                Console.WriteLine("Chyba: Poskodeny alebo neexistujuci konfiguracny subor scsConfig.xml");
                return;
            }

            if (args.Length > 0)
            {
                //DateTime debug_start = DateTime.Now;
                
                try
                {
                    caRootCertPublicKey = X509Certificate.FromDER(BIO.File(caRootCertPublicKeyFile,"r")).PublicKey;
                }
                catch
                {
                    Console.WriteLine("Chyba: Neplatny alebo neexistujuci korenovy certifikat");
                    return;
                }
        
                if ((args[0] == "--zmena-certifikatu")||(args[0] == "-z"))               
                {
                    if (args[1] == null)
                    {
                        Console.WriteLine("Chybny parameter\n\nDostupne parametre:\n --zmena-certifikatu (-z) <heslo aktualneho certifikatu SiS>\n --prenos-na-scs (-p)");
                        return;
                    }
                    activeCertPass = args[1];

                    try
                    {
                        temporaryCertificate = CryptoKey.FromPrivateKey(BIO.File(temporaryCertificateFile, "r"), activeCertPass);
                    }
                    catch
                    {
                        Console.WriteLine("Chyba: Neplatny alebo neexistujuci aktualny certifikat SiS");
                        return;
                    }

                    activeCertificatesSN = new int[totalSisCertificateNumber];
                    try
                    {
                        for (int i = totalSisCertificateNumber; i > 0; i--)
                        {
                            X509Certificate cert = X509Certificate.FromDER(BIO.File("Certs/" + i.ToString() + ".der", "r"));
                            activeCertificatesSN[totalSisCertificateNumber - i] = cert.SerialNumber;                            
                        }
                    }
                    catch
                    {
                        Console.WriteLine("Chyba: Neboli najdene vsetky certifikaty SiS");
                        return;
                    }

                    string path = "IncomingVotes/";
                    try
                    {
                        voteInputSchema = XDocument.Load(voteInputSchemaFile);
                    }
                    catch
                    {
                        Console.WriteLine("Chyba: Neplatna XML schema pre 3. obalku s casovou peciatkou");
                        return;
                    }
                    try
                    {
                        voteDecryptedSchema = XDocument.Load(voteDecryptedSchemaFile);
                    }
                    catch
                    {
                        Console.WriteLine("Chyba: Neplatna XML schema pre 2. obalku");
                        return;
                    }
                    try
                    {
                        voteDecryptedWithTimeStampSchema = XDocument.Load(voteDecryptedWithTimeStampSchemaFile);
                    }
                    catch
                    {
                        Console.WriteLine("Chyba: Neplatna XML schema pre 2. obalku s casovou peciatkou");
                        return;
                    }

                    DirectoryInfo incomingVotesDirectory = new DirectoryInfo(path);
                    int incomingVotesNumber = incomingVotesDirectory.GetFileSystemInfos("*.xml").Length;
                    string randomVotePath;
                    XDocument randomVote = null;
                    string randomVoteCertSN;
                    System.Security.Cryptography.SHA512 hashSHA512 = System.Security.Cryptography.SHA512.Create();


                    for (int i = 0; i < incomingVotesNumber; i++)
                    {
                        randomVotePath = getRandomFile(path, ".xml");
                        try
                        {
                            randomVote = XDocument.Parse(File.ReadAllText(randomVotePath));
                            if (!Validate_XML(randomVote, voteInputSchema))
                                continue;

                            randomVoteCertSN = randomVote.XPathSelectElement("/voteBallotThirdEnvelope/temporaryTOSSCertificateSerialNumber").Value;
                        }
                        catch
                        {
                            randomVoteCertSN = "-1";
                        }

                        try
                        {
                            if (int.Parse(randomVoteCertSN) == activeCertificatesSN[0])
                            {
                                // Certificate S/N is actual - process vote - phase 1 (decipher)
                                XElement randomVoteTS = randomVote.XPathSelectElement("/timestampType");
                                randomVote = decipherVote(randomVote, temporaryCertificate);
                                if (!Validate_XML(randomVote, voteDecryptedWithTimeStampSchema))
                                {
                                    File.AppendAllText("votesForTransfer/falseVotesTimestampList.toss", randomVoteTS.XPathSelectElement("//serialNumber").Value + " " + randomVoteTS.XPathSelectElement("//dataHash").Value + " " + randomVoteTS.XPathSelectElement("//dateTime").Value + "\n");
                                    continue;
                                }

                                if (verifyVote(randomVote))
                                {
                                    String tsSN = randomVote.XPathSelectElement("/voteBallotSecondEnvelope/timestamp/serialNumber").Value;

                                    X509Certificate voterSignCert;
                                    BIO voterCertBio = BIO.MemoryBuffer();
                                    voterCertBio.Write(Convert.FromBase64String(randomVote.XPathSelectElement("/voteBallotSecondEnvelope/voterSignCertificate").Value));
                                    voterSignCert = X509Certificate.FromDER(voterCertBio);

                                    String randomVoteVoterCertSN = voterSignCert.SerialNumber.ToString();
                                    File.AppendAllText("VotesStorage/" + randomVoteVoterCertSN + ".tsl", Convert.ToBase64String(hashSHA512.ComputeHash(Encoding.ASCII.GetBytes(randomVote.XPathSelectElement("/voteBallotSecondEnvelope/voteBallot").Value))) + " " + tsSN + " " + i.ToString() + "\n");
                                    String voteSavePath = "VotesStorage/" + i.ToString() + ".xml"; //"-" + randomVoteCertSN + "-" + tsSN +
                                    randomVote.Save(voteSavePath);
                                    File.Delete(randomVotePath);
                                }
                                else
                                {
                                    String voteSavePath = "InvalidVotesForTransfer/" + i.ToString() + ".xml";
                                    randomVote.Save(voteSavePath);
                                    File.Delete(randomVotePath);
                                }
                            }
                            else if (activeCertificatesSN.Contains(int.Parse(randomVoteCertSN)))
                            {
                                // To be processed in next rounds
                                File.Move(randomVotePath, path + "WaitForNextRound/" + randomVotePath.Substring(randomVotePath.LastIndexOf('\\') + 1));
                            }
                            else
                            {
                                // Wrong certificate
                                File.Move(randomVotePath, "InvalidVotesForTransfer/" + randomVotePath.Substring(randomVotePath.LastIndexOf('\\') + 1));
                            }
                        }
                        catch
                        {
                            continue;
                        }
                    }

                        try
                        {
                           
                            XDocument config = XDocument.Load(configFile);
                            config.XPathSelectElement("/duhConfig/totalSisCertificateNumber").Value = (totalSisCertificateNumber - 1).ToString();
                            config.Save(configFile);
                        }
                        catch
                        {
                            Console.WriteLine("Upozornenie: Nie je mozne zmenit polozku totalSisCertificateNumber v duhConfig.xml. Zmenu je potrebne vykonat rucne.");
                        }

                    // Move back votes for future decryption
                    DirectoryInfo waitForNextRoundDirectory = new DirectoryInfo(path+"WaitForNextRound/");
                    FileInfo[] nextRoundFiles = waitForNextRoundDirectory.GetFiles();
                    foreach (FileInfo nextRoundFile in nextRoundFiles)
                    {
                        nextRoundFile.MoveTo(nextRoundFile.Directory.Parent.FullName + "/" + nextRoundFile.Name);
                    }
                    
                }

                else if ((args[0] == "--prenos-na-scs")||(args[0] == "-p"))
                {
                    try
                    {
                        voteDecryptedWithTimeStampSchema = XDocument.Load(voteDecryptedWithTimeStampSchemaFile);
                    }
                    catch
                    {
                        Console.WriteLine("Chyba: Neplatna XML schema pre 2. obalku s casovou peciatkou");
                        return;
                    }

                    //subor s nazvom s/n cert. volica.tsl - zoznam vsetkych s/n casovych peciatok (po skonceni spracovavania nacitat do pamati/stromu, usporiadat a zmazat duplikaty)
                    string path = "VotesStorage/";
                    string pathTransfer = "VotesForTransfer/";
                    string pathInvalid = "InvalidVotesForTransfer/";
                    string[] voteTSList;
                    string randomVotesTSListPath;


                    DateTime time = DateTime.UtcNow;
                    DirectoryInfo incomingVotesDirectory = new DirectoryInfo(path);
                    int votesTSListNumber = incomingVotesDirectory.GetFileSystemInfos("*.tsl").Length;
                    for (int i = 0; i < votesTSListNumber; i++)
                    {
                        randomVotesTSListPath = getRandomFile(path, ".tsl");
                        try
                        {                            
                            voteTSList = File.ReadAllLines(randomVotesTSListPath, Encoding.ASCII);
                            Array.Sort<string>(voteTSList);

                            string previousHash = "";
                            string actualHash = "";
                            int storedTSSerialNumber = int.MaxValue;
                            int storedFileNumber = -1;
                            int actualBlockTSSerialNumber = int.MaxValue;
                            int actualBlockFileNumber = -1;
                            int actualTSSerialNumber = -1;
                            int actualFileNumber = -1;
                            foreach (string sort in voteTSList)
                            {
                                //prejst cele pole voteTSList - pre kazdy rovnaky hash nacitat S/N TS zo stringu medzi medzerami
                                //pamatat si najvacsie cislo S/N TS a k nemu cislo subora, ale ignorovat S/N TS pre opakovane hashe hlasov
                                //presunut dany jediny platny hlas po anonymizovani do VotesForTransfer

                                previousHash = actualHash;
                                actualHash = sort.Substring(0, 88);
                                actualTSSerialNumber = int.Parse(sort.Substring(sort.IndexOf(' ') + 1, sort.LastIndexOf(' ') - sort.IndexOf(' ') - 1));
                                actualFileNumber = int.Parse(sort.Substring(sort.LastIndexOf(' ') + 1));

                                if (actualHash == previousHash)
                                {
                                    if (actualTSSerialNumber <= actualBlockTSSerialNumber)
                                    {
                                        actualBlockTSSerialNumber = actualTSSerialNumber;
                                        actualBlockFileNumber = actualFileNumber;
                                    }
                                }
                                else
                                {
                                    if (actualBlockTSSerialNumber <= storedTSSerialNumber)
                                    {
                                        storedTSSerialNumber = actualBlockTSSerialNumber;
                                        storedFileNumber = actualBlockFileNumber;

                                        actualBlockTSSerialNumber = actualTSSerialNumber;
                                        actualBlockFileNumber = actualFileNumber;
                                    }
                                }
                            }

                            if (actualBlockTSSerialNumber <= storedTSSerialNumber)
                            {
                                storedTSSerialNumber = actualBlockTSSerialNumber;
                                storedFileNumber = actualBlockFileNumber;
                            }

                        
                            string certPath = randomVotesTSListPath.Substring(randomVotesTSListPath.LastIndexOf('\\')+1);
                            certPath = certPath.Remove(certPath.LastIndexOf('.'));

                            try
                            {
                                XDocument vote = XDocument.Load(path + storedFileNumber.ToString() + ".xml"); //"-" + certPath + "-" + storedTSSerialNumber.ToString() +
                                if (!Validate_XML(vote, voteDecryptedWithTimeStampSchema))
                                    continue;
                                vote.XPathSelectElement("/voteBallotSecondEnvelope/voteBallot").Save(pathTransfer + i.ToString() + ".xml");
                                File.Delete(path + storedFileNumber.ToString() + ".xml");
                            }
                            catch
                            {
                                try
                                {
                                    File.Delete(randomVotesTSListPath);
                                }
                                catch
                                {
                                    continue;
                                }
                                continue;
                            }
                            
                            File.Delete(randomVotesTSListPath);                            
                        }
                        catch
                        {
                        }
                    }

                    //presun vsetkych suborov do InvalidVotesForTransfer
                    DirectoryInfo invalidVotesDirectory = new DirectoryInfo(pathInvalid);
                    int invalidCount = invalidVotesDirectory.GetFileSystemInfos("*.xml").Length;
                    DirectoryInfo remainingVotesDirectory = new DirectoryInfo(path);
                    int remainingCount = remainingVotesDirectory.GetFileSystemInfos("*.xml").Length;
                    string remainingFile;
                    for(int j = 0; j < remainingCount; j++)
                    {
                        remainingFile = getRandomFile(path, ".xml");
                        File.SetCreationTimeUtc(remainingFile, time);
                        File.SetLastAccessTimeUtc(remainingFile, time);
                        File.SetLastWriteTimeUtc(remainingFile, time);
                        File.Move(remainingFile, pathInvalid + (invalidCount + j).ToString() + ".xml");
                    }
                    Directory.Delete(path,true);
                    Directory.CreateDirectory(path);
                }
                else
                {
                    Console.WriteLine("Neznamy parameter\n\nDostupne parametre:\n --zmena-certifikatu (-z) <heslo aktualneho certifikatu SiS>\n --prenos-na-scs (-p)");
                }

                //DateTime debug_end = DateTime.Now;
                //Console.WriteLine((debug_end.Subtract(debug_start).TotalSeconds / 1000).ToString()+ "\n" + debug_end.Subtract(debug_start).ToString());
            }
            else
            {
                Console.WriteLine("Nezadany parameter\n\nDostupne parametre:\n --zmena-certifikatu (-z) <heslo aktualneho certifikatu SiS>\n --prenos-na-scs (-p)");
            }
        }
    }
}
