﻿using System;
using System.IO;
using System.Windows.Forms;
using OpenSSL.Core;
using OpenSSL.Crypto;
using OpenSSL.X509;

namespace jednoduchaCertifikacnaAutorita
{
    public partial class Form1 : Form
    {
        string certPath = "cert/";
        public Form1()
        {
            InitializeComponent();

            //Pouzite pre testovanie vykonnosti
            /*DateTime debug_start = DateTime.Now;
            for (int i = 0; i < 1000; i++)
            {                
                certificateSubjectName.Text = "Test " + i.ToString();
                button2_Click(this, null);
            }
            DateTime debug_end = DateTime.Now;
            MessageBox.Show((debug_end.Subtract(debug_start).TotalSeconds / 2000).ToString(), debug_end.Subtract(debug_start).ToString());*/
        }

        private void button1_Click(object sender, EventArgs e)
        {  
            BigNumber bn = 0x10001;
            RSA eVoteSign = new RSA();
            eVoteSign.GenerateKeys(4096, bn, null, null);            
            CryptoKey key = new CryptoKey(eVoteSign);
            eVoteSign.Dispose();
   
            X509V3ExtensionList extList = new X509V3ExtensionList();
            extList.Add(new X509V3ExtensionValue("subjectKeyIdentifier", false, "hash"));
            extList.Add(new X509V3ExtensionValue("authorityKeyIdentifier", false, "keyid:always,issuer:always"));
            extList.Add(new X509V3ExtensionValue("basicConstraints", true, "critical,CA:true"));
            extList.Add(new X509V3ExtensionValue("keyUsage", false, "cRLSign,keyCertSign"));
  
            using (X509CertificateAuthority certificateAutority = X509CertificateAuthority.SelfSigned(
                new SimpleSerialNumber(),
                key,
                MessageDigest.SHA512,
                "e-Vote Root Certificate Authority",
                DateTime.Now,
                TimeSpan.FromDays(2922),
                extList))
            {
                X509Certificate rootCA = certificateAutority.Certificate;
                File.WriteAllBytes(certPath+"rootCA.der", rootCA.DER);                
                File.WriteAllText(certPath+"maximumserialnumber.rca", rootCA.SerialNumber.ToString());
                BIO outputPrivateKey = BIO.File(certPath + "rootCA.pem", "w");
                key.WritePrivateKey(outputPrivateKey, Cipher.AES_256_CBC, "heslo");
                outputPrivateKey.SetClose(BIO.CloseOption.Close);
                outputPrivateKey.Dispose();
            }
            
        }

        private void button2_Click(object sender, EventArgs e)
        {
            BigNumber bn = 0x10001;
            RSA eVoteSign = new RSA();
            eVoteSign.GenerateKeys(4096, bn, null, null);
            CryptoKey key = new CryptoKey(eVoteSign);
            eVoteSign.Dispose();
            Configuration cfg = null;
            
            if (radioButton1.Checked)
            {                
                cfg = new Configuration("certDigSign.cnf");
            }
            else if (radioButton2.Checked)
            {                
                cfg = new Configuration("certEncryption.cnf");
            }
            
            X509Certificate rootCA = X509Certificate.FromDER(BIO.File(certPath+"rootCA.DER","r"));            
            int rootCAmaximumserialnumber = int.Parse(File.ReadAllText(certPath+"maximumserialnumber.rca"));
            X509CertificateAuthority rootACA = new X509CertificateAuthority(rootCA, CryptoKey.FromPrivateKey(BIO.File(certPath + "rootCA.pem","r"), "heslo"), new SimpleSerialNumber(rootCAmaximumserialnumber), cfg);
            if (certificateSubjectName.Text.Length == 0)
            {
                MessageBox.Show("Chyba: Nie je zadané meno vlastníka certifikátu", "Chyba");
                return;
            }
            X509Request newRCA = new X509Request(3, certificateSubjectName.Text, key);
            X509Certificate newCA = rootACA.ProcessRequest(newRCA, DateTime.Now, DateTime.Now.AddYears(1),MessageDigest.SHA512);

            File.WriteAllText(certPath + "maximumserialnumber.rca", (rootCAmaximumserialnumber + 1).ToString());
            File.WriteAllBytes(certPath + (rootCAmaximumserialnumber+1).ToString() + ".DER", newCA.DER);
            BIO outputPrivateKey = BIO.File(certPath + (rootCAmaximumserialnumber + 1).ToString() + ".pem", "w");
            key.WritePrivateKey(outputPrivateKey, Cipher.AES_256_CBC, "heslo");
            outputPrivateKey.SetClose(BIO.CloseOption.Close);
            outputPrivateKey.Dispose();           
               
        }

        private void textBox1_TextChanged(object sender, EventArgs e)
        {

        }

        private bool verifyCertificate(X509Certificate inputCert, CryptoKey certAuthorityKey)
        {
            //neuplne - overuje iba hash a podpis hashu
            //neoveruje cas a platnost voci CRL
            return inputCert.Verify(certAuthorityKey);
        }

        private void button3_Click(object sender, EventArgs e)
        {
            if (openFileDialog1.ShowDialog() == DialogResult.OK)
            {
                X509Certificate rootCA = X509Certificate.FromDER(BIO.File(certPath + "rootCA.der", "r"));
                X509Certificate validatedCert = X509Certificate.FromDER(BIO.File(openFileDialog1.FileName, "r"));
                              
                //overenie platnosti podpisu cert voci rootCA
                MessageBox.Show(verifyCertificate(validatedCert,rootCA.PublicKey).ToString());
            }
        }

        private void Form1_Load(object sender, EventArgs e)
        {

        }
    }
}
