﻿using System;
using System.IO;
using System.Text;
using System.Xml;
using System.Xml.Linq;
using System.Xml.Schema;
using System.Xml.XPath;
using OpenSSL.Core;
using OpenSSL.Crypto;

namespace ScitaciServer
{
    class Program
    {
        static string configFile = "scsConfig.xml";

        static string voteDecryptedSchemaFile;
        static string voteEncryptedSchemaFile;
        static string voteInputSchemaFile;
        static string candidateListSchemaFile;
        static string evotingCert;
        static string evotingCertPass;
        static string votesDir;
        static string invalidVotesDir;
        static string resultsFile;
        static string processedNIHFile;
        static string candidateListFile;

        static XDocument voteDecryptedSchema;
        static XDocument voteInputSchema;
        static XDocument candidateListSchema;
        static XDocument voteEncryptedSchema;

        public static bool Validate_XML(XDocument validated_xml, XDocument xsd_schema)
        {
            bool no_errors = true;
            try
            {
                XmlSchemaSet schemas = new XmlSchemaSet();
                schemas.Add(null, XmlReader.Create(new StringReader(xsd_schema.ToString())));
                validated_xml.Validate(schemas, null, true);
            }
            catch
            {
                no_errors = false;
            }
            return no_errors;
        }

        static XDocument decipherVote(XDocument vote, CryptoKey cert)
        {
            byte[] secondEnvelopeEnc = Convert.FromBase64String(vote.XPathSelectElement("/voteBallotFirstEnvelope/voteBallotEncrypted").Value);
            byte[] secondEnvelopeEncKey = Convert.FromBase64String(vote.XPathSelectElement("/voteBallotFirstEnvelope/voteBallotEncryptionKey").Value);
            RSA envelopeDecryption = cert.GetRSA();
            byte[] decData = envelopeDecryption.PrivateDecrypt(secondEnvelopeEncKey, RSA.Padding.OAEP);

            CipherContext symetricCipher = new CipherContext(Cipher.AES_256_CBC);
            byte[] iv = new byte[8];
            Buffer.BlockCopy(decData, 0, iv, 0, 8);
            byte[] symKey = new byte[256];
            Buffer.BlockCopy(decData, 8, symKey, 0, 256);
            byte[] secondEnvelope = symetricCipher.Decrypt(secondEnvelopeEnc, symKey, iv);
            XDocument result = XDocument.Parse(Encoding.ASCII.GetString(secondEnvelope));
            return result;
        }


        static void Main(string[] args)
        {
            Console.WriteLine("Scitaci server (ScS)");
            try
            {
                XDocument config = XDocument.Load(configFile);
                voteDecryptedSchemaFile = config.XPathSelectElement("/scsConfig/voteDecryptedSchemaFile").Value.Trim();
                voteEncryptedSchemaFile = config.XPathSelectElement("/scsConfig/voteEncryptedSchemaFile").Value.Trim();
                voteInputSchemaFile = config.XPathSelectElement("/scsConfig/voteInputSchemaFile").Value.Trim();
                candidateListSchemaFile = config.XPathSelectElement("/scsConfig/candidateListSchemaFile").Value.Trim();
                evotingCert = config.XPathSelectElement("/scsConfig/evotingKey").Value.Trim();
                votesDir = config.XPathSelectElement("/scsConfig/votesDir").Value.Trim();
                invalidVotesDir = config.XPathSelectElement("/scsConfig/invalidVotesDir").Value.Trim();
                resultsFile = config.XPathSelectElement("/scsConfig/resultsFile").Value.Trim();
                processedNIHFile = config.XPathSelectElement("/scsConfig/processedNIHFile").Value.Trim();
                candidateListFile = config.XPathSelectElement("/scsConfig/candidateListFile").Value.Trim();                
            }
            catch
            {
                Console.WriteLine("Chyba: Poskodeny alebo neexistujuci konfiguracny subor scsConfig.xml");
                return;
            }

            if (args.Length > 0)
            {                
                if ((args[0] == "--spracuj-hlasy")||(args[0] == "-s"))
                {
                    if(args.Length != 2)
                    {
                        Console.WriteLine("Nezadany parameter\n\nDostupne parametre:\n --spracuj-hlasy (-s) <heslo pre elektronicke volby>");
                        return;
                    }
                    evotingCertPass = args[1];

                    try
                    {
                        candidateListSchema = XDocument.Load(candidateListSchemaFile);
                    }
                    catch
                    {
                        Console.WriteLine("Chyba: Neplatna XML schema pre zoznam kandidatov");
                        return;
                    }

                    try
                    {
                        voteInputSchema = XDocument.Load(voteInputSchemaFile);
                    }
                    catch
                    {
                        Console.WriteLine("Chyba: Neplatna XML schema pre vstupny subor z DUH");
                        return;
                    }

                    try
                    {
                        voteEncryptedSchema = XDocument.Load(voteEncryptedSchemaFile);
                    }
                    catch
                    {
                        Console.WriteLine("Chyba: Neplatna XML schema pre zasifrovany hlas");
                        return;
                    }

                    try
                    {
                        voteDecryptedSchema = XDocument.Load(voteDecryptedSchemaFile);
                    }
                    catch
                    {
                        Console.WriteLine("Chyba: Neplatna XML schema pre hlasovaci listok");
                        return;
                    }

                    XDocument candidateList = XDocument.Load(candidateListFile);
                    if (!Validate_XML(candidateList, candidateListSchema))
                    {
                        Console.WriteLine("Chyba: Neplatny format zoznamu kandidatov");
                        return;
                    }

                    //DateTime debug_start = DateTime.Now;

                    CryptoKey evotingCertificate = CryptoKey.FromPrivateKey(BIO.File(evotingCert, "r"), evotingCertPass);
                    DirectoryInfo di = new DirectoryInfo(votesDir);
                    FileInfo[] rgFiles = di.GetFiles("*.xml");

                    int candidateCount = Convert.ToInt32((double)(candidateList.XPathEvaluate("count(/candidateList/candidate)")) + 1);                    
                    int[] evotingResult = new int[candidateCount];
                    for(int i = 0; i < candidateCount; i++)
                        evotingResult[i] = 0;
                    XDocument vote;

                    //spracovanie platnych hlasov
                    //DateTime debug_start = DateTime.Now;

                    int candidateChoice;
                    string dateOfVotesCount = DateTime.Now.ToString();                    
                    File.AppendAllText(resultsFile, "Vysledky volieb zo dna " + dateOfVotesCount + ":\n");
                    
                    foreach (FileInfo voteFile in rgFiles)
                    {
                        try
                        {
                            vote = XDocument.Load(voteFile.FullName);

                            //overenie xml formatu
                            if (!Validate_XML(vote, voteInputSchema))
                                continue;

                            vote = XDocument.Parse(Encoding.ASCII.GetString(Convert.FromBase64String(vote.XPathSelectElement("/voteBallot").Value)));
                            if (!Validate_XML(vote, voteEncryptedSchema))
                                continue;

                            vote = decipherVote(vote, evotingCertificate);
                            if (!Validate_XML(vote, voteDecryptedSchema))
                                continue;
                            candidateChoice = int.Parse(vote.XPathSelectElement("/voteBallot/selectedChoice/index").Value);

                            if (candidateChoice < candidateCount)
                                evotingResult[candidateChoice]++;
                            File.AppendAllText(processedNIHFile, vote.XPathSelectElement("/voteBallot/randomVoteIdentifier").Value + "\n");
                        }
                        catch
                        {
                            evotingResult[0]++;
                            continue;
                        }
                    }
                    //DateTime debug_end = DateTime.Now;
                    //Console.WriteLine(debug_end.Subtract(debug_start).ToString());
                    //spracovanie NIH z neplatnych hlasov
                    DirectoryInfo di2 = new DirectoryInfo(invalidVotesDir);
                    FileInfo[] rgInvFiles = di2.GetFiles("*.xml");

                    foreach (FileInfo invalidVoteFile in rgInvFiles)
                    {
                        try
                        {
                            vote = XDocument.Load(invalidVoteFile.FullName);
                            if (!Validate_XML(vote, voteInputSchema))
                                continue;
                            vote = XDocument.Parse(Encoding.ASCII.GetString(Convert.FromBase64String(vote.XPathSelectElement("/voteBallot").Value)));
                            if (!Validate_XML(vote, voteEncryptedSchema))
                                continue;
                            vote = decipherVote(vote, evotingCertificate);
                            if (!Validate_XML(vote, voteDecryptedSchema))
                                continue;
                            File.AppendAllText(processedNIHFile, vote.XPathSelectElement("/voteBallot/randomVoteIdentifier").Value + "\n");
                        }
                        catch
                        {
                            continue;
                        }
                    }

                    //usporiadanie NIH a odstranenie duplikatov
                    string[] processedNIH = File.ReadAllLines(processedNIHFile);
                    Array.Sort<string>(processedNIH);
                    File.Delete(processedNIHFile);
                    string prevString = "";
                    foreach (string processedNIHstring in processedNIH)
                    {
                        if (prevString != processedNIHstring)
                        {
                            File.AppendAllText(processedNIHFile, processedNIHstring + "\n");
                            prevString = processedNIHstring;
                        }
                    }
                    File.AppendAllText(processedNIHFile, "Zoznam spracovanych NIH zo dna " + dateOfVotesCount + ".\n");
                    File.AppendAllText(resultsFile, "Neplatnych hlasov: " + evotingResult[0].ToString() + "\n");                    
                    for (int i = 1; i < candidateCount; i++)
                        File.AppendAllText(resultsFile, candidateList.XPathSelectElement("/candidateList/candidate[@id = "+i.ToString()+"]/name").Value.Trim() + ": " + evotingResult[i].ToString() + "\n");

                    //DateTime debug_end = DateTime.Now;
                    //Console.WriteLine((debug_end.Subtract(debug_start).TotalSeconds / 1000).ToString() + "\n" + debug_end.Subtract(debug_start).ToString());
                }
                else
                {
                    Console.WriteLine("Neznamy parameter\n\nDostupne parametre:\n --spracuj-hlasy (-s) <heslo pre elektronicke volby>");
                }
            }
            else
            {
                Console.WriteLine("Nezadany parameter\n\nDostupne parametre:\n --spracuj-hlasy (-s) <heslo pre elektronicke volby>");
            }
        }
    }
}
