﻿using System;
using System.IO;
using System.Linq;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Threading;
using System.Xml;
using System.Xml.Linq;
using System.Xml.Schema;
using System.Xml.XPath;
using OpenSSL.Core;
using OpenSSL.Crypto;
using OpenSSL.X509;
using Org.BouncyCastle.Crypto.Engines;
using Org.BouncyCastle.Crypto.Parameters;
using Org.BouncyCastle.Math;
using Org.BouncyCastle.Security;

namespace SietovyServer
{
    class Program
    {
        private static string getRandomFile(string path)
        {
            string file = null;
            if (!string.IsNullOrEmpty(path))
            {
                var extensions = new string[] { ".xml" };
                try
                {
                    var di = new DirectoryInfo(path);                    
                    var rgFiles = di.GetFiles("*.*").Where(f => extensions.Contains(f.Extension.ToLower()));
                    SecureRandom R = new SecureRandom();
                    file = rgFiles.ElementAt(R.Next(0, rgFiles.Count())).FullName;
                }
                catch { }
            }
            return file;
        }

        public static void Main(string[] args)
        {
            if (args.Length == 0)
            {
                Console.WriteLine("Neznamy parameter\n\nDostupne parametre:\n--spustit-volby (-s) <heslo pre elektronicky podpis NIH>\n--zmenit-certifikat (-z)");
                return;
            }
            else
            {
                if ((args[0] == "--spustit-volby") || (args[0] == "-s"))
                {
                    if (args.Length == 2)
                    {
                        Server SietovyServer = new Server(args[1]);
                    }
                    else
                    {
                        Console.WriteLine("Neznamy parameter\n\nDostupne parametre:\n--spustit-volby (-s) <heslo pre elektronicky podpis NIH>\n--zmenit-certifikat (-z)");
                        return;
                    }
                }
                else if ((args[0] == "--zmenit-certifikat") || (args[0] == "-z"))
                {
                    //DateTime debug_start = DateTime.Now;

                    bool tryAgain = false;
                    int maxSavedVotesNumber;
                    StreamReader maxSavedNumReadStream = null;
                    FileStream maxSavedVotesNumFile = null;
                    int minSavedVotesNumber;
                    FileStream minSavedVotesNumFile;
                    StreamReader minSavedNumReadStream;


                    while (!tryAgain)
                    {
                        tryAgain = false;
                        try
                        {
                            maxSavedVotesNumFile = File.Open("Data/maxSavedVotes.toss", FileMode.Open, FileAccess.Read, FileShare.None);
                            maxSavedNumReadStream = new StreamReader(maxSavedVotesNumFile);
                        }
                        catch
                        {
                            tryAgain = true;
                        }
                    }

                    try
                    {
                        maxSavedVotesNumber = int.Parse(maxSavedNumReadStream.ReadLine()) + 1;
                    }
                    catch
                    {
                        Console.WriteLine("Chyba: neplatny obsah suboru maxSavedVotes.toss");
                        maxSavedNumReadStream.Close();
                        return;
                    }

                    maxSavedNumReadStream.Close();

                    try
                    {
                        minSavedVotesNumFile = File.Open("Data/minSavedVotes.toss", FileMode.Open, FileAccess.ReadWrite, FileShare.None);
                        minSavedNumReadStream = new StreamReader(minSavedVotesNumFile);
                        minSavedVotesNumber = int.Parse(minSavedNumReadStream.ReadLine());
                    }
                    catch
                    {
                        Console.WriteLine("Chyba: neplatny obsah suboru minSavedVotes.toss");
                        maxSavedNumReadStream.Close();
                        return;
                    }

                    StreamWriter minSavedNumWriteStream = new StreamWriter(minSavedVotesNumFile);
                    try
                    {
                        minSavedVotesNumFile.SetLength(0);
                        minSavedNumWriteStream.Write(maxSavedVotesNumber);
                    }
                    catch
                    {
                        Console.WriteLine("Chyba: subor minSavedVotes.toss je chraneny proti zapisu");
                        minSavedNumWriteStream.Close();
                        return;
                    }

                    minSavedNumWriteStream.Close();

                    string path;
                    DateTime time = DateTime.UtcNow;
                    for (int i = minSavedVotesNumber; i < maxSavedVotesNumber; i++)
                    {
                        path = "votesForTransfer/" + i.ToString() + ".xml";
                        try
                        {
                            File.Move(getRandomFile("votes"), path);
                            File.SetCreationTimeUtc(path, time);
                            File.SetLastAccessTimeUtc(path, time);
                            File.SetLastWriteTimeUtc(path, time);
                        }
                        catch
                        {
                            Console.WriteLine("Upozornenie: subor " + path + " neexistuje");
                        }
                    }

                    //DateTime debug_end = DateTime.Now;
                    //Console.WriteLine((debug_end.Subtract(debug_start).TotalSeconds / 1000).ToString() +"\n"+ debug_end.Subtract(debug_start).ToString());
                }
                else
                {
                    Console.WriteLine("Neznamy parameter\n\nDostupne parametre:\n--spustit-volby (-s) <heslo pre elektronicky podpis NIH>\n--zmenit-certifikat (-z)");
                    return;
                }
            }
        }
    }


  class Server
  {
      static string configFile = "sisConfig.xml";

      int onlineServerPort;
      IPAddress onlineTimeStampAddress;
      int onlineTimeStampPort;
      string blindSignatureCertificateFile;
      string blindSignaturePrivateKeyFile;

      static string voteInputSchemaFile;
      static string timestampSchemaFile;

    private TcpListener tcpListener;
    private Thread listenThread;
    private X509Certificate blindSignatureCertificate;
    private CryptoKey blindSignaturePrivateKey;

    static RsaKeyParameters blindSignKey;

    public static bool Validate_XML(XDocument validated_xml, XDocument xsd_schema)
    {
        bool no_errors = true;
        try
        {
            XmlSchemaSet schemas = new XmlSchemaSet();
            schemas.Add(null, XmlReader.Create(new StringReader(xsd_schema.ToString())));
            validated_xml.Validate(schemas, null, true);
        }
        catch
        {
            no_errors = false;
        }
        return no_errors;
    }

    private static string UnescapeXML(string s)
    {
        if (string.IsNullOrEmpty(s)) return s;

        string returnString = s;
        returnString = returnString.Replace("&apos;", "'");
        returnString = returnString.Replace("&quot;", "\"");
        returnString = returnString.Replace("&gt;", ">");
        returnString = returnString.Replace("&lt;", "<");
        returnString = returnString.Replace("&amp;", "&");

        return returnString;
    }

    private XDocument GetTimeStamp(XDocument vote)
    {
        Socket timestampSocket;
        XDocument voteResponse;
        System.Security.Cryptography.SHA512 hashSHA512 = System.Security.Cryptography.SHA512.Create();
        byte[] voteHash = hashSHA512.ComputeHash(Encoding.ASCII.GetBytes(vote.ToString()));
        
        try
        {
            timestampSocket = sendConnectionSendPhase(onlineTimeStampAddress, onlineTimeStampPort, voteHash);
        }
        catch
        {
            Console.WriteLine(DateTime.Now.ToString() + " Chyba: nie je mozne nadviazat spojenie s Poskytovatelom sluzby casovej peciatky");
            return null;
        }

        byte[] response = sendConnectionResponsePhase(timestampSocket);
        if (response == null)
        {
            Console.WriteLine(DateTime.Now.ToString() + " Chyba: Poskytovatel sluzby casovej peciatky neodpoveda");
            return null;
        }

        try
        {
            voteResponse = XDocument.Parse(Encoding.ASCII.GetString(response));
        }
        catch
        {
            return null;
        }

        return voteResponse;
    }

    public static BigInteger opensslBigToBigInteger(String opensslInt)
    {
        String hexInt;
        try
        {
            hexInt = opensslInt.Replace(":", "");
            hexInt = hexInt.Replace("\\s+", "");
            return new BigInteger(hexInt, 16);
        }
        catch
        {
            return null;
        }
    }

    private byte[] MyBCSign(byte[] subject)//, RSA rsa)
    {
        RsaEngine rsaBlindSignEngine = new RsaEngine();
        byte[] signedData;
        //RsaKeyParameters blindSignKey = new RsaKeyParameters(true, opensslBigToBigInteger(rsa.PublicModulus.ToString()), opensslBigToBigInteger(rsa.PrivateExponent.ToString()));
        try
        {
            rsaBlindSignEngine.Init(true, blindSignKey);
            signedData = rsaBlindSignEngine.ProcessBlock(subject, 0, subject.Length);
        }
        catch
        {
            Console.WriteLine(DateTime.Now.ToString() + " Chyba: Podpis NIH nie je mozne vytvorit");
            return null;
        }
        return signedData;
    }

    private XElement SignNIH(string nih)
    {

        try
        {
            byte[] data = Convert.FromBase64String(nih);

            byte[] signedData = MyBCSign(data);//, blindSignaturePrivateKey.GetRSA());

            //toss = temporaryOnlineStorageServer (old name for NetworkServer/SietovyServer)
            XElement blindSignatureEnvelope = new XElement("signatureNIH",
                                        new XElement("nihData", Convert.ToBase64String(data)),
                                        new XElement("tossSignCertificate", Convert.ToBase64String(blindSignatureCertificate.DER)),
                                        new XElement("nihSignature", Convert.ToBase64String(signedData))
                                        );
            return blindSignatureEnvelope;
        }
        catch
        {
            return null;
        }
    }

    public Server(string blindSignaturePrivateKeyPassword)
    {
        try
        {
            XDocument config = XDocument.Load(configFile);
            onlineServerPort = int.Parse(config.XPathSelectElement("/sisConfig/onlineServerPort").Value.Trim());
            onlineTimeStampAddress = IPAddress.Parse(config.XPathSelectElement("/sisConfig/onlineTimeStampAddress").Value.Trim());
            voteInputSchemaFile = config.XPathSelectElement("/sisConfig/voteInputSchemaFile").Value.Trim();
            timestampSchemaFile = config.XPathSelectElement("/sisConfig/timestampSchemaFile").Value.Trim();
            onlineTimeStampPort = int.Parse(config.XPathSelectElement("/sisConfig/onlineTimeStampPort").Value.Trim());
            blindSignatureCertificateFile = config.XPathSelectElement("/sisConfig/blindSignatureCertificateFile").Value.Trim();
            blindSignaturePrivateKeyFile = config.XPathSelectElement("/sisConfig/blindSignaturePrivateKeyFile").Value.Trim();            
        }
        catch
        {
            Console.WriteLine("Chyba: Poskodeny alebo neexistujuci konfiguracny subor sisConfig.xml");
            return;
        }

        try
        {
            blindSignatureCertificate = X509Certificate.FromDER(BIO.File(blindSignatureCertificateFile, "r"));
        }
        catch
        {
            Console.WriteLine("Chyba: Poskodeny alebo neexistujuci certifikat pre overenie podpisu NIH");
            return;
        }

        try
        {
            blindSignaturePrivateKey = CryptoKey.FromPrivateKey(BIO.File(blindSignaturePrivateKeyFile, "r"), blindSignaturePrivateKeyPassword);
        }
        catch
        {
            Console.WriteLine("Chyba: Poskodeny alebo neexistujuci sukromny kluc (alebo heslo) pre podpis NIH");
            return;
        }

        try
        {
            blindSignKey = new RsaKeyParameters(true, opensslBigToBigInteger(blindSignaturePrivateKey.GetRSA().PublicModulus.ToString()), opensslBigToBigInteger(blindSignaturePrivateKey.GetRSA().PrivateExponent.ToString()));
        }
        catch
        {
            Console.WriteLine("Chyba: Poskodeny sukromny kluc (alebo heslo) pre podpis NIH");
            return;
        }

      try
      {
          this.tcpListener = new TcpListener(IPAddress.Any, onlineServerPort);
          this.listenThread = new Thread(new ThreadStart(ListenForClients));
          this.listenThread.Start();
      }
      catch
      {
          Console.WriteLine("Chyba: Port " + onlineServerPort.ToString() + " nie je dostupny");
          return;
      }
    }

    private void ListenForClients()
    {
        this.tcpListener.Start();
        Console.WriteLine("Sietovy Server (SiS)");

        while (true)
        {
            try
            {
                TcpClient client = this.tcpListener.AcceptTcpClient();
                Thread clientThread = new Thread(new ParameterizedThreadStart(HandleClientComm));
                clientThread.Start(client);
            }
            catch {}
        }
    }       

       private byte[] receiveConnectionReceivePhase(NetworkStream clientStream)
       {
           byte[] receivedData = new byte[0];
           byte[] message = new byte[1024];
           byte[] concat;
           int bytesRead;

           while (true)
           {
               bytesRead = 0;

               try
               {
                   //blocks until a client sends a message                
                   bytesRead = clientStream.Read(message, 0, 1024);
                   concat = new byte[receivedData.Length + bytesRead];
                   Buffer.BlockCopy(receivedData, 0, concat, 0, receivedData.Length);
                   Buffer.BlockCopy(message, 0, concat, receivedData.Length, bytesRead);
                   receivedData = concat;
               }
               catch
               {
                   //a socket error has occured
                   //Console.WriteLine("Voting Application - connection error");
                   receivedData = null;
                   break;
               }

               if (bytesRead == 0)
               {
                   //the client has disconnected from the server
                   //Console.WriteLine("Voting Application - disconnected/waiting for response");
                   break;
               }


               //message has successfully been received
           }
           return receivedData;
       }

       private void receiveConnectionResponsePhase(NetworkStream clientStream, byte[] data)
       {
           clientStream.Write(data, 0, data.Length);
       }

       private Socket sendConnectionSendPhase(IPAddress ip_adress, int port, byte[] sendData)
       {
           var clientSocket = new Socket(AddressFamily.InterNetwork,
                                          SocketType.Stream, ProtocolType.Tcp);
           clientSocket.Connect(ip_adress, port);
           clientSocket.Send(sendData);
           clientSocket.Disconnect(false);
           return clientSocket;
       }

       private byte[] sendConnectionResponsePhase(Socket clientSocket)
       {
           var buffer = new byte[256];
           byte[] responseMessage = new byte[0];
           byte[] concat;
           int bytesRead;
           while (true)
           {
               try
               {
                   bytesRead = clientSocket.Receive(buffer);
               }
               catch
               {
                   //Console.WriteLine("Chyba: spojenie prerusene");
                   return null;
               }
               concat = new byte[responseMessage.Length + bytesRead];
               Buffer.BlockCopy(responseMessage, 0, concat, 0, responseMessage.Length);
               Buffer.BlockCopy(buffer, 0, concat, responseMessage.Length, bytesRead);
               responseMessage = concat;
               if (bytesRead == 0)
               {
                   break;
               }
           }
           clientSocket.Dispose();
           return responseMessage;
       }
           
    private void HandleClientComm(object client)
    {
        //Console.WriteLine("Voting Application - connected");
        byte[] voteMessage;
        TcpClient tcpClient;
        NetworkStream clientStream;
        XDocument vote;
        XDocument voteInputSchema;
        XDocument timestampSchema;
        //DateTime debug_start = DateTime.Now;
        try
        {
            voteInputSchema = XDocument.Load(voteInputSchemaFile);
        }
        catch
        {
            Console.WriteLine(DateTime.Now.ToString() + " Chyba: Neplatna XML schema pre 3. obalku");
            return;
        }

        try
        {
            timestampSchema = XDocument.Load(timestampSchemaFile);
        }
        catch
        {
            Console.WriteLine(DateTime.Now.ToString() + " Chyba: Neplatna XML schema pre casovu peciatku");
            return;
        }

        try
        {
            tcpClient = (TcpClient)client;
            clientStream = tcpClient.GetStream();

            voteMessage = receiveConnectionReceivePhase(clientStream);
        }
        catch
        {
            Console.WriteLine(DateTime.Now.ToString() + " Chyba: Spojenie s volebnou aplikaciou zlyhalo");
            return;
        }

        if ((voteMessage != null) && (voteMessage.Length > 0))
        {
            try
            {
                StringReader sr = new StringReader(UnescapeXML(Encoding.ASCII.GetString(voteMessage)));
                vote = XDocument.Load(sr);
            }
            catch
            {
                tcpClient.Close();
                return;
            }

            if (!Validate_XML(vote,voteInputSchema))
            {
                tcpClient.Close();
                return;
            }

            //sparsovanie hlasu - pridanie TS, slepy podpis
            XElement nih = vote.XPathSelectElement("//randomVoteIdentifier");
            vote.XPathSelectElement("//randomVoteIdentifier").Remove();

            nih = SignNIH(nih.Value);
            if (nih == null)
            {
                tcpClient.Close();
                return;
            }

            XDocument timestamp = GetTimeStamp(vote);
            if (!Validate_XML(timestamp, timestampSchema))
            {
                tcpClient.Close();
                return;
            }

            XDocument voteResponse = new XDocument(
                                        new XElement("voteResponse",
                                            timestamp.Root,
                                            nih));         

            //ulozenie hlasu na dalsie spracovanie
            bool tryAgain = false;
            StreamReader maxSavedNumReadStream = null;
            FileStream maxSavedVotesNumFile = null;
            int maxSavedVotesNumber;
            while (!tryAgain)
            {
                tryAgain = false;
                try
                {
                    maxSavedVotesNumFile = File.Open("Data/maxSavedVotes.toss", FileMode.Open, FileAccess.ReadWrite, FileShare.None);
                    maxSavedNumReadStream = new StreamReader(maxSavedVotesNumFile);
                }
                catch
                {
                    tryAgain = true;
                }
            }

            try
            {
                maxSavedVotesNumber = int.Parse(maxSavedNumReadStream.ReadLine()) + 1;
            }
            catch
            {
                Console.WriteLine(DateTime.Now.ToString() + " Chyba: Subor maxSavedVotes.toss je poskodeny");
                tcpClient.Close();
                return;
            }

            try
            {
                maxSavedVotesNumFile.SetLength(0);
                StreamWriter maxSavedNumWriteStream = new StreamWriter(maxSavedVotesNumFile);
                maxSavedNumWriteStream.Write(maxSavedVotesNumber);
                maxSavedNumWriteStream.Close();
            }
            catch
            {
                Console.WriteLine(DateTime.Now.ToString() + " Chyba: Subor maxSavedVotes.toss je nedostupny");
                tcpClient.Close();
                return;
            }

            vote.Root.LastNode.AddAfterSelf(timestamp.Root);
            try
            {
                vote.Save("votes/" + (maxSavedVotesNumber.ToString()) + ".xml");
            }
            catch
            {
                Console.WriteLine(DateTime.Now.ToString() + " Chyba: Chyba pri ukladani hlasu");
                tcpClient.Close();
                return;
            }


            //odoslanie odpovede - TS + podpisany NIH   
            try
            {
                receiveConnectionResponsePhase(clientStream, Encoding.ASCII.GetBytes(voteResponse.ToString()));
            }
            catch {}
            //DateTime debug_end = DateTime.Now;
            //Console.WriteLine((debug_end.Subtract(debug_start).TotalSeconds / 1000).ToString() + "\n" + debug_end.Subtract(debug_start).ToString());
        }

        tcpClient.Close();
    }
  }
}
            
            