﻿using System;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Windows.Forms;
using System.Xml;
using System.Xml.Linq;
using System.Xml.Schema;
using System.Xml.XPath;
using OpenSSL.Core;
using OpenSSL.Crypto;
using OpenSSL.X509;
using Org.BouncyCastle.Crypto.Digests;
using Org.BouncyCastle.Crypto.Engines;
using Org.BouncyCastle.Crypto.Generators;
using Org.BouncyCastle.Crypto.Parameters;
using Org.BouncyCastle.Crypto.Signers;
using Org.BouncyCastle.Math;
using Org.BouncyCastle.Security;

namespace VolebnaAplikacia
{   
    public partial class Form1 : Form
    {
        string configFile = "vaConfig.xml";

        string candidatesListSchemaFile;
        string voteResponseSchemaFile;
        string outputVoteFileName;
        IPAddress networkServerAddress;
        string temporaryDirectoryPath;
        int publishingServerTimeout;
        string candidatesListUrl;
        string firstEnvelopeEncryptionCert;
        string thirdEnvelopeEncryptionCert;
        string blindSignatureOnlineServerCert;
        int networkServerPort;
        
        private PasswordForm passForm = new PasswordForm();
        XPathDocument candidatesList;
        string candidatesListFileName;
        XDocument vote;
        XDocument voteResponseSchema;

        //DEBUG OFF
        //static int i;
        
        private byte[] voteSign(byte[] subject, RSA rsa)
        {
            byte[] result;
            try
            {
                RsaEngine rsaEngine = new RsaEngine();
                RsaKeyParameters signKey = new RsaKeyParameters(true, opensslBigToBigInteger(rsa.PublicModulus.ToString()), opensslBigToBigInteger(rsa.PrivateExponent.ToString()));
                PssSigner voteSigner = new PssSigner(rsaEngine, new Sha512Digest());
                voteSigner.Init(true, signKey);
                voteSigner.BlockUpdate(subject, 0, subject.Length);
                result = voteSigner.GenerateSignature();
            }
            catch
            {
                return null;
            }
            return result;
        }

        private Socket sendConnectionSendPhase(IPAddress ip_adress, int port, byte[] sendData)
        {
            var clientSocket = new Socket(AddressFamily.InterNetwork,
                                           SocketType.Stream, ProtocolType.Tcp);
            clientSocket.Connect(ip_adress, port);
            clientSocket.Send(sendData);
            clientSocket.Disconnect(false);
            return clientSocket;
        }

        private byte[] sendConnectionResponsePhase(Socket clientSocket)
        {
            var buffer = new byte[256];
            byte[] responseMessage = new byte[0];
            byte[] concat;
            int bytesRead;
            while (true)
            {
                try
                {
                    bytesRead = clientSocket.Receive(buffer);
                }
                catch
                {
                    MessageBox.Show("Chyba: Spojenie prerušené","Chyba");
                    return null;
                }
                concat = new byte[responseMessage.Length + bytesRead];
                Buffer.BlockCopy(responseMessage, 0, concat, 0, responseMessage.Length);
                Buffer.BlockCopy(buffer, 0, concat, responseMessage.Length, bytesRead);
                responseMessage = concat;
                if (bytesRead == 0)
                {
                    break;
                }
            }
            clientSocket.Dispose();
            return responseMessage;
        }
        
        public static BigInteger opensslBigToBigInteger(String opensslInt)
        {
            try
            {
                String hexInt = opensslInt.Replace(":", "");
                hexInt = hexInt.Replace("\\s+", "");
                return new BigInteger(hexInt, 16);
            }
            catch
            {
                return null;
            }
        }    

        public Form1()
        { 
            InitializeComponent();
            buttonLoadCandidatesList.Visible = true;
            buttonShowCandidatesList.Visible = false;
            buttonCreateVote.Visible = false;
            buttonSendVote.Visible = false;
            richTextBox1.Visible = false;

            try
            {
                XDocument config = XDocument.Load(configFile);
                candidatesListSchemaFile = config.XPathSelectElement("/vaConfig/candidatesListSchemaFile").Value.Trim();
                voteResponseSchemaFile = config.XPathSelectElement("/vaConfig/voteResponseSchemaFile").Value.Trim();
                outputVoteFileName = config.XPathSelectElement("/vaConfig/outputVoteFileName").Value.Trim();
                networkServerAddress = IPAddress.Parse(config.XPathSelectElement("/vaConfig/temporaryOnlineStorageServerAddress").Value.Trim());
                temporaryDirectoryPath = config.XPathSelectElement("/vaConfig/temporaryDirectoryPath").Value.Trim();
                publishingServerTimeout = int.Parse(config.XPathSelectElement("/vaConfig/publishingServerTimeout").Value.Trim());
                candidatesListUrl = config.XPathSelectElement("/vaConfig/candidatesListUrl").Value.Trim();
                firstEnvelopeEncryptionCert = config.XPathSelectElement("/vaConfig/firstEnvelopeEncryptionCert").Value.Trim();
                thirdEnvelopeEncryptionCert = config.XPathSelectElement("/vaConfig/thirdEnvelopeEncryptionCert").Value.Trim();
                blindSignatureOnlineServerCert = config.XPathSelectElement("/vaConfig/blindSignatureOnlineServerCert").Value.Trim();
                networkServerPort = int.Parse(config.XPathSelectElement("/vaConfig/temporaryOnlineStorageServerPort").Value.Trim());
            }
            catch
            {
                MessageBox.Show("Chyba: Poskodeny alebo neexistujuci konfiguracny subor vaConfig.xml");
                return;
            }
            //richTextBox1.AppendText("Voting Application has started");            
        }

        public static bool SaveFileFromURL(string url, string destinationFileName, int timeoutInSeconds)
        {
            // Create a web request to the URL
            HttpWebRequest MyRequest = (HttpWebRequest)WebRequest.Create(url);
            MyRequest.Timeout = timeoutInSeconds * 1000;
            try
            {
                // Get the web response
                HttpWebResponse MyResponse = (HttpWebResponse)MyRequest.GetResponse();

                // Make sure the response is valid
                if (HttpStatusCode.OK == MyResponse.StatusCode)
                {
                    // Open the response stream
                    using (Stream MyResponseStream = MyResponse.GetResponseStream())
                    {
                        // Open the destination file
                        using (FileStream MyFileStream = new FileStream(destinationFileName, FileMode.Create, FileAccess.Write))
                        {
                            // Create a 4K buffer to chunk the file
                            byte[] MyBuffer = new byte[4096];
                            int BytesRead;
                            // Read the chunk of the web response into the buffer
                            while (0 < (BytesRead = MyResponseStream.Read(MyBuffer, 0, MyBuffer.Length)))
                            {
                                // Write the chunk from the buffer to the file
                                MyFileStream.Write(MyBuffer, 0, BytesRead);
                            }
                        }
                    }
                }
            }
            catch //(Exception err)
            {
                return false;
            }
            return true;
        }

        private void showCandidatesList_click(object sender, EventArgs e)
        {
            if (candidatesList != null)
            {                
                XPathNavigator candidateListNavigator = candidatesList.CreateNavigator();
                XPathNodeIterator candidateListIterator = candidateListNavigator.Select("candidateList/candidate");
                candidateListSelector.Items.Add("Prázdny hlas");
                //richTextBox1.AppendText("\n--- Waiting for voter's choice");
                while (candidateListIterator.MoveNext())
                {
                    candidateListSelector.Items.Add(candidateListIterator.Current.SelectSingleNode("name").Value);
                }
                candidateListSelector.SelectedIndex = 0;
                candidateListSelector.Visible = true;
                if (inputNIH.Text.Length == 20)
                    buttonCreateVote.Visible = true;
            }
        }

        private void Form1_Load(object sender, EventArgs e)
        {
            try
            {
                voteResponseSchema = XDocument.Load(voteResponseSchemaFile);
            }
            catch
            {
                MessageBox.Show("Chyba: Neplatna XML schema pre odpoved volebneho servera","Chyba");
                return;
            }
            loadCandidatesList_click(this, null);

            //DEBUG OFF
        /*    DateTime debug_start = DateTime.Now;
            System.Random ran = new System.Random(DateTime.Now.Millisecond);            
            for (i = 1; i < 1001; i++)
            {
                //candidateListSelector.SelectedIndex = ran.Next(6);
                //generateNIH_Click(this,null);
                //createVote_Click(this, null);
                //buttonSendVote_Click(this, null);
                button2_Click(this, null);
            }
            DateTime debug_end = DateTime.Now;
            MessageBox.Show((debug_end.Subtract(debug_start).TotalSeconds / 1000).ToString(), debug_end.Subtract(debug_start).ToString());
            */
        }

        private void loadCandidatesList_click(object sender, EventArgs e)
        {
            candidatesListFileName = temporaryDirectoryPath + "candidateList.xml";
            if (!SaveFileFromURL(candidatesListUrl, candidatesListFileName, publishingServerTimeout))
            {
                MessageBox.Show("Chyba: Publikačný server nie je dostupný.", "Chyba");
                Application.Exit();
            }

            //richTextBox1.AppendText("\n--- Validating " + candidatesListFileName + " against " + xsd_candidatesList);
            if (Validate_XML_File(candidatesListFileName, candidatesListSchemaFile))
                    {                        
                        candidatesList = new XPathDocument(candidatesListFileName);
                        buttonLoadCandidatesList.Visible = false;
                        //richTextBox1.AppendText("\n--- Candidate list " + candidatesListFileName + " loaded");
                        showCandidatesList_click(sender,null);
                    }
                    else
                    {
                        MessageBox.Show("Chyba: Zoznam kandidátov je poškodený.", "Chyba");
                        Application.Exit();                        
                    }              
        }

        //Validacia XML suboru (adresa) voci XML Schema (adresa)
        public static bool Validate_XML_File(string xml, string xsd)
        {            
            bool no_errors = true;
            try
            {
                XDocument validated_xml = XDocument.Load(xml);
                XDocument xsd_schema = XDocument.Load(xsd);
                XmlSchemaSet schemas = new XmlSchemaSet();
                schemas.Add(null, XmlReader.Create(new StringReader(xsd_schema.ToString())));
                validated_xml.Validate(schemas, null, true);
            }
            catch
            {
                no_errors = false;
            }
            return no_errors;
        }

        public static bool Validate_XML_Loaded(XDocument validated_xml, XDocument xsd_schema)
        {
            bool no_errors = true;
            try
            {
                XmlSchemaSet schemas = new XmlSchemaSet();
                schemas.Add(null, XmlReader.Create(new StringReader(xsd_schema.ToString())));
                validated_xml.Validate(schemas, null, true);
            }
            catch
            {
                no_errors = false;
            }
            return no_errors;
        }

        private XDocument createFirstEnvelope(XDocument vote, X509Certificate eVotingCertificate)
        {
            //TODO: Verify cert. 
            
            try
            {
                RSA firstEnvelopeEncryption = eVotingCertificate.PublicKey.GetRSA();
                CipherContext symetricCipher = new CipherContext(Cipher.AES_256_CBC);
                byte[] symKey = new byte[256];
                SecureRandom.GetInstance("SHA1PRNG").NextBytes(symKey);
                byte[] iv = new byte[8];
                SecureRandom.GetInstance("SHA1PRNG").NextBytes(iv);

                System.Text.ASCIIEncoding encoding = new System.Text.ASCIIEncoding();
                byte[] encdata = new byte[iv.Length + symKey.Length];
                Buffer.BlockCopy(iv, 0, encdata, 0, 8);
                Buffer.BlockCopy(symKey, 0, encdata, 8, 256);

                XElement firstEnvelope = new XElement("voteBallotFirstEnvelope",
                                           new XElement("voteBallotEncrypted", Convert.ToBase64String(symetricCipher.Encrypt(encoding.GetBytes(vote.ToString()), symKey, iv))),
                                           new XElement("voteBallotEncryptionKey", Convert.ToBase64String(firstEnvelopeEncryption.PublicEncrypt(encdata, RSA.Padding.OAEP)))  
                                           );
                XDocument result = new XDocument();
                result.AddFirst(firstEnvelope);
                return result;
            }
            catch 
            {
                return null; 
            }
        }

        private XDocument createSecondEnvelope(XDocument vote, X509Certificate voterSignCertificate, CryptoKey voterPrivateKey)
        {
            try
            {
            RSA secondEnvelopeSignature = voterPrivateKey.GetRSA();
            byte[] data = Encoding.ASCII.GetBytes(vote.ToString());            
            
            XElement secondEnvelope = new XElement("voteBallotSecondEnvelope",
                                            new XElement("voteBallot", Convert.ToBase64String(data)),
                                            new XElement("voterSignCertificate", Convert.ToBase64String(voterSignCertificate.DER)),                    
                                            new XElement("voteBallotSignature", Convert.ToBase64String(voteSign(data, secondEnvelopeSignature)))
                                            );           
                XDocument result = new XDocument();
                result.AddFirst(secondEnvelope);
                return result;
            }
            catch
            {
                return null;
            }
        }

        private XDocument createThirdEnvelope(XDocument vote, X509Certificate eVotingTemporaryStorageCertificate)
        {
            //TODO: Verify cert.
            try
            {                
                RSA thirdEnvelopeEncryption = eVotingTemporaryStorageCertificate.PublicKey.GetRSA();
                CipherContext symetricCipher = new CipherContext(Cipher.AES_256_CBC);
                byte[] symKey = new byte[256];
                SecureRandom.GetInstance("SHA1PRNG").NextBytes(symKey);
                byte[] iv = new byte[8];
                SecureRandom.GetInstance("SHA1PRNG").NextBytes(iv);                   

                System.Text.ASCIIEncoding encoding = new System.Text.ASCIIEncoding();
                byte[] encdata = new byte[iv.Length+symKey.Length];
                Buffer.BlockCopy(iv, 0, encdata, 0, 8);
                Buffer.BlockCopy(symKey, 0, encdata, 8, 256);

                // toss = temporaryOnlineStorageServer (old name for NetworkServer/SietovyServer)
                XElement thirdEnvelope = new XElement("voteBallotThirdEnvelope",
                                            new XElement("voteBallotEncrypted", Convert.ToBase64String(symetricCipher.Encrypt(encoding.GetBytes(vote.ToString()),symKey,iv))),
                                            new XElement("voteBallotEncryptionKey", Convert.ToBase64String(thirdEnvelopeEncryption.PublicEncrypt(encdata, RSA.Padding.OAEP))),
                                            new XElement("temporaryTOSSCertificateSerialNumber", eVotingTemporaryStorageCertificate.SerialNumber),
                                            new XElement("randomVoteIdentifier", inputNIH.Text)
                                            );
                XDocument result = new XDocument();
                result.AddFirst(thirdEnvelope);
                return result;
            }
            catch
            {                
                return null;
            }
        }        

        private void createVote_Click(object sender, EventArgs e)
        {
            buttonCreateVote.Enabled = false;
            button1.Enabled = false;
            inputNIH.Enabled = false;
            candidateListSelector.Enabled = false;

            XElement voteElement = new XElement("voteBallot",
                                        new XElement("selectedChoice",
                                            new XElement("index", candidateListSelector.SelectedIndex),
                                            new XElement("name", candidateListSelector.SelectedItem)
                                            ),
                                        new XElement("randomVoteIdentifier", inputNIH.Text)
                                        );

            if (!SaveFileFromURL(firstEnvelopeEncryptionCert, temporaryDirectoryPath + "firstEnvelopeCert.cer", publishingServerTimeout))
            {
                MessageBox.Show("Chyba: Publikačný server nie je dostupný. Hlas nebol vytvorený", "Chyba");
                buttonCreateVote.Enabled = true;
                button1.Enabled = true;
                inputNIH.Enabled = true;
                candidateListSelector.Enabled = true;
                return;
            }
            if (!SaveFileFromURL(thirdEnvelopeEncryptionCert, temporaryDirectoryPath + "thirdEnvelopeCert.cer", publishingServerTimeout))
            {
                MessageBox.Show("Chyba: Publikačný server nie je dostupný. Hlas nebol vytvorený", "Chyba");
                buttonCreateVote.Enabled = true;
                button1.Enabled = true;
                inputNIH.Enabled = true;
                candidateListSelector.Enabled = true;
                return;
            }
            vote = new XDocument();
            
                    openFileDialog3.Title = "Zvoľte certifikát na tvorbu elektronického podpisu";
                    openFileDialog3.Filter = "Certifikáty | *.DER";
                    if (openFileDialog3.ShowDialog() == DialogResult.OK)
                    {                        
                        if (passForm.ShowDialog() == DialogResult.OK)
                        {
                            vote.AddFirst(voteElement);                            
                            try
                            { 
                            vote = createThirdEnvelope(createSecondEnvelope(createFirstEnvelope(vote, X509Certificate.FromDER(BIO.File(temporaryDirectoryPath + "firstEnvelopeCert.cer", "r"))), X509Certificate.FromDER(BIO.File(openFileDialog3.FileName, "r")), CryptoKey.FromPrivateKey(BIO.File(openFileDialog3.FileName.Substring(0, openFileDialog3.FileName.Length-3) + "pem", "r"), passForm.password)), X509Certificate.FromDER(BIO.File(temporaryDirectoryPath + "thirdEnvelopeCert.cer", "r")));
                                //DEBUG OFF
                                //vote = createThirdEnvelope(createSecondEnvelope(createFirstEnvelope(vote, X509Certificate.FromDER(BIO.File(temporaryDirectoryPath + "firstEnvelopeCert.cer", "r"))), X509Certificate.FromDER(BIO.File(i.ToString() + ".der", "r")), CryptoKey.FromPrivateKey(BIO.File(i.ToString() + ".pem", "r"), "heslo")), X509Certificate.FromDER(BIO.File(temporaryDirectoryPath + "thirdEnvelopeCert.cer", "r")));
                                vote.Save(outputVoteFileName);
                                buttonSendVote.Visible = true;
                                buttonCreateVote.Enabled = false;
                                button1.Enabled = false;
                                inputNIH.Enabled = false;
                                candidateListSelector.Enabled = false;
                           }
                            catch
                            {
                                MessageBox.Show("Chyba: Hlas nebol vytvorený. Zlé heslo?", "Chyba");
                                buttonCreateVote.Enabled = true;
                                button1.Enabled = true;
                                inputNIH.Enabled = true;
                                candidateListSelector.Enabled = true;
                            }

                        }
                    }
        }

        private void generateNIH_Click(object sender, EventArgs e)
        {
            byte[] rnd = new byte[256];
            new SecureRandom().NextBytes(rnd);
            System.Security.Cryptography.SHA512 hash = System.Security.Cryptography.SHA512.Create();
            inputNIH.Text = Convert.ToBase64String(hash.ComputeHash(rnd)).Substring(0, 20);
            if (buttonLoadCandidatesList.Visible == false)
                buttonCreateVote.Visible = true;
        }

        private void buttonSendVote_Click(object sender, EventArgs e)
        {    
            //uprava randomVoteIdentifier vo vote pre potreby blindSign            
            String dataNIH = vote.XPathSelectElement("/voteBallotThirdEnvelope/randomVoteIdentifier").Value;
            //zapis cert. z netu
            if (!SaveFileFromURL(blindSignatureOnlineServerCert, temporaryDirectoryPath + "blindSignatureCert.cer", publishingServerTimeout))
            {
                MessageBox.Show("Chyba: Publikačný server nie je dostupný. Hlas nebol vytvorený", "Chyba");
                return;
            }

            RSA blindScheme;
            RsaKeyParameters publicKey;
            try
            {
                blindScheme = X509Certificate.FromDER(BIO.File(temporaryDirectoryPath + "blindSignatureCert.cer", "r")).PublicKey.GetRSA();
                publicKey = new RsaKeyParameters(false, opensslBigToBigInteger(blindScheme.PublicModulus.ToString()), opensslBigToBigInteger(blindScheme.PublicExponent.ToString()));
            }
            catch
            {
                MessageBox.Show("Chyba: Interná chyba aplikácie", "Chyba");
                return;
            }
       
            byte[] msg = Encoding.ASCII.GetBytes(dataNIH);
            
            RsaBlindingFactorGenerator blindFactorGen = new RsaBlindingFactorGenerator();
            blindFactorGen.Init(publicKey);            
            BigInteger blindFactor = blindFactorGen.GenerateBlindingFactor();
            RsaBlindingEngine blindingEngine = new RsaBlindingEngine();
            RsaBlindingParameters parameters = new RsaBlindingParameters(publicKey, blindFactor);            
            PssSigner blindSigner = new PssSigner(blindingEngine, new Sha512Digest());
            
            // vytvorenie blind signature obalky
            byte[] salt = SecureRandom.GetSeed(20);
            ParametersWithRandom param = new ParametersWithRandom(parameters, new SecureRandom(salt));
            blindSigner.Init(true, param);            
            blindSigner.BlockUpdate(msg, 0, msg.Length);
            byte[] blindedData = blindSigner.GenerateSignature();

            vote.XPathSelectElement("/voteBallotThirdEnvelope/randomVoteIdentifier").Value = Convert.ToBase64String(blindedData);
           
            //odoslanie hlasu
            Socket clientSocket = sendConnectionSendPhase(networkServerAddress, networkServerPort, Encoding.ASCII.GetBytes(vote.ToString()));
            byte[] responseMessage = sendConnectionResponsePhase(clientSocket);
            XDocument responseMessageXML;

            try
            {
                responseMessageXML = XDocument.Parse(Encoding.ASCII.GetString(responseMessage));
            }
            catch
            {
                MessageBox.Show("Chyba: Neplatná odpoveď volebného servera", "Chyba");
                return;
            }

            if (!Validate_XML_Loaded(responseMessageXML, voteResponseSchema))
            {
                MessageBox.Show("Chyba: Neplatná odpoveď volebného servera", "Chyba");
                return;
            }

            responseMessageXML.XPathSelectElement("/voteResponse/signatureNIH/nihData").Value = Convert.ToBase64String(Encoding.ASCII.GetBytes(dataNIH));
            byte[] blindedSig = Convert.FromBase64String(responseMessageXML.XPathSelectElement("/voteResponse/signatureNIH/nihSignature").Value);
            
            // odstranenie blind signature obalky            
            blindingEngine.Init(false, parameters);
            byte[] signatureValue = blindingEngine.ProcessBlock(blindedSig, 0, blindedSig.Length);
                   
            responseMessageXML.XPathSelectElement("/voteResponse/signatureNIH/nihSignature").Value = Convert.ToBase64String(signatureValue);

            //overenie vrateneho podpisaneho NIH + TS a ulozenie
            //TODO: overit platnost TS
                        
            PssSigner Verifier = new PssSigner(new RsaEngine(), new Sha512Digest());
            Verifier.Init(false, publicKey);
            Verifier.BlockUpdate(msg, 0, msg.Length);

            if (Verifier.VerifySignature(signatureValue))
            {
                MessageBox.Show("NIH: " + Encoding.ASCII.GetString(Convert.FromBase64String(responseMessageXML.XPathSelectElement("//nihData").Value)) + "\nTS S/N: " + responseMessageXML.XPathSelectElement("//timestamp/serialNumber").Value);
                saveFileDialog1.OverwritePrompt = true;
                saveFileDialog1.Filter = "eVote súbory (*.evote)| *.evote";
                if (saveFileDialog1.ShowDialog() == DialogResult.OK)
                {
                    responseMessageXML.Save(saveFileDialog1.FileName);
                }
            }
            else
            {
                MessageBox.Show("Chyba: Neplatný podpis NIH volebným serverom", "Chyba");
            } 
         
        }

        private void button2_Click(object sender, EventArgs e)
        {
            vote = XDocument.Load("temp/lastSentVote.xml");
            buttonSendVote_Click(sender, e);
        }

        private static string UnescapeXML(string s)
        {
            if (string.IsNullOrEmpty(s)) return s;

            string returnString = s;
            returnString = returnString.Replace("&apos;", "'");
            returnString = returnString.Replace("&quot;", "\"");
            returnString = returnString.Replace("&gt;", ">");
            returnString = returnString.Replace("&lt;", "<");
            returnString = returnString.Replace("&amp;", "&");

            return returnString;
        }

        private void inputNIH_TextChanged(object sender, EventArgs e)
        {
            if (inputNIH.TextLength == 20)
            {
                buttonCreateVote.Visible = true;
            }
            else
            {
                buttonCreateVote.Visible = false;
            }
        }

        
    }
}
